mod args;
mod assert;
mod binread_temp;
mod c_enum;
mod calc;
mod custom_writer;
mod endian;
mod r#enum;
mod if_cond;
mod ignore;
mod import;
mod magic;
mod map;
mod map_stream;
mod padding;
mod restore_position;
mod simple;
mod stream;
mod struct_generic;
mod top_level_map;
